/* stconfig.h.in.  Generated from configure.in by autoheader.  */

/* Define if you want to use internal GSM library */
#undef ENABLE_GSM

/* Define if you want to use fast ALAW conversions */
#undef FAST_ALAW_CONVERSION

/* Define if you want to use fast ULAW conversions */
#undef FAST_ULAW_CONVERSION

/* Define if you have ALSA installed */
#undef HAVE_ALSA

/* Define if you have ALSA 0.4 installed */
#undef HAVE_ALSA4

/* Define if you have ALSA 0.5 installed */
#undef HAVE_ALSA5

/* Define if you have ALSA 0.9 installed */
#undef HAVE_ALSA9

/* Define to 1 if you have the <byteswap.h> header file. */
#undef HAVE_BYTESWAP_H

/* Define to 1 if you have the <errno.h> header file. */
#define HAVE_ERRNO_H 1

/* Define to 1 if you have the <fcntl.h> header file. XXX */
#define HAVE_FCNTL_H 1

/* Define to 1 if you have the `getopt' function. */
#undef HAVE_GETOPT

/* Define to 1 if you have the <getopt.h> header file. */
#undef HAVE_GETOPT_H

/* Define to 1 if you have the <inttypes.h> header file. */
#undef HAVE_INTTYPES_H

/* Define if you have LAME (LAME Ain't an MP3 Encoder) library installed */
#undef HAVE_LAME

/* Define to 1 if you have the `m' library (-lm). */
#undef HAVE_LIBM

/* Define if you have MAD (MP3 Audio Decoder) Library installed */
#undef HAVE_LIBMAD

/* Define if you have Ogg Vorbis Library installed */
#undef HAVE_LIBVORBIS

/* Define to 1 if you have the <limits.h> header file. XXX*/
#define HAVE_LIMITS_H 1

/* Define to 1 if you have the <linux/asound.h> header file. */
#undef HAVE_LINUX_ASOUND_H

/* Define to 1 if you have the <machine/soundcard.h> header file. */
#undef HAVE_MACHINE_SOUNDCARD_H

/* Define to 1 if you have the <malloc.h> header file. XXX*/
#define HAVE_MALLOC_H 1

/* Define to 1 if you have the `memmove' function. */
#define HAVE_MEMMOVE 1

/* Define to 1 if you have the <memory.h> header file. */
#undef HAVE_MEMORY_H

/* Define if you have OSS installed */
#undef HAVE_OSS

/* Define to 1 if you have the `rand' function. */
#undef HAVE_RAND

/* Define to 1 if you have the <sound/asound.h> header file. */
#undef HAVE_SOUND_ASOUND_H

/* Define to 1 if you have the <stdint.h> header file. */
#undef HAVE_STDINT_H

/* Define to 1 if you have the <stdlib.h> header file. */
#define HAVE_STDLIB_H 1

/* Define to 1 if you have the `strcasecmp' function. */
#undef HAVE_STRCASECMP

/* Define to 1 if you have the `strdup' function. XXX*/
#define HAVE_STRDUP 1

/* Define to 1 if you have the `strerror' function. */
#define HAVE_STRERROR 1

/* Define to 1 if you have the <strings.h> header file. */
#undef HAVE_STRINGS_H

/* Define to 1 if you have the <string.h> header file. XXX*/
#define HAVE_STRING_H 1

/* Define if you have Sun /dev/audio installed */
#undef HAVE_SUNAUDIO

/* Define to 1 if you have the <sun/audioio.h> header file. */
#undef HAVE_SUN_AUDIOIO_H

/* Define to 1 if you have the <sys/audioio.h> header file. */
#undef HAVE_SYS_AUDIOIO_H

/* Define to 1 if you have the <sys/soundcard.h> header file. */
#undef HAVE_SYS_SOUNDCARD_H

/* Define to 1 if you have the <sys/stat.h> header file. */
#define HAVE_SYS_STAT_H 1

/* Define to 1 if you have the <sys/types.h> header file. */
#define HAVE_SYS_TYPES_H 1

/* Define to 1 if you have the <unistd.h> header file. */
#undef HAVE_UNISTD_H

/* Define to the address where bug reports for this package should be sent. */
#undef PACKAGE_BUGREPORT

/* Define to the full name of this package. */
#undef PACKAGE_NAME

/* Define to the full name and version of this package. */
#undef PACKAGE_STRING

/* Define to the one symbol short name of this package. */
#undef PACKAGE_TARNAME

/* Define to the version of this package. */
#undef PACKAGE_VERSION

/* The number of bytes in type char */
#define SIZEOF_CHAR 1

/* The number of bytes in type int */
#define SIZEOF_INT 4

/* The number of bytes in type long */
#define SIZEOF_LONG 4

/* The number of bytes in type short */
#define SIZEOF_SHORT 2

/* The number of bytes in type void* */
#define SIZEOF_VOIDP 4

/* Define to 1 if you have the ANSI C header files. */
#undef STDC_HEADERS

/* Define if you want to use old rate routines */
#undef USE_OLD_RATE

/* Define to 1 if your processor stores words with the most significant byte
   first (like Motorola and SPARC, unlike Intel and VAX). */
#undef WORDS_BIGENDIAN

/* Define to empty if `const' does not conform to ANSI C. */
#undef const

/* Define as `__inline' if that's what the C compiler calls it, or to nothing
   if it is not supported. */
#define inline __inline

/* Define to `unsigned' if <sys/types.h> does not define. */
#undef size_t

#define stat _stat
