typedef int int32_t;
typedef unsigned int uint32_t;
typedef int int32_t;
typedef unsigned int uint32_t;
typedef unsigned char uint8_t;
typedef char int8_t;
typedef unsigned short uint16_t;
typedef short int16_t;
typedef unsigned __int64 uint64_t;
typedef __int64 int64_t;

#define __STDC__ 1